// Copyright (c) 2015 Microsoft Corporation. All rights reserved.
// Released under Apache 2.0 license as described in the file LICENSE.
// DO NOT EDIT, automatically generated file, generator scripts/get_tokens_cpp.py
#include "util/name.h"
namespace lean {
void initialize_tokens();
void finalize_tokens();
name const & get_aliases_tk();
name const & get_period_tk();
name const & get_backtick_tk();
name const & get_dslash_tk();
name const & get_fieldarrow_tk();
name const & get_placeholder_tk();
name const & get_colon_tk();
name const & get_semicolon_tk();
name const & get_dcolon_tk();
name const & get_orelse_tk();
name const & get_lparen_tk();
name const & get_rparen_tk();
name const & get_llevel_curly_tk();
name const & get_lcurly_tk();
name const & get_rcurly_tk();
name const & get_ldcurly_tk();
name const & get_rdcurly_tk();
name const & get_lbracket_tk();
name const & get_rbracket_tk();
name const & get_lcurlybang_tk();
name const & get_rcurlybang_tk();
name const & get_langle_tk();
name const & get_rangle_tk();
name const & get_triangle_tk();
name const & get_caret_tk();
name const & get_do_tk();
name const & get_up_tk();
name const & get_bar_tk();
name const & get_comma_tk();
name const & get_add_tk();
name const & get_sub_tk();
name const & get_greater_tk();
name const & get_question_tk();
name const & get_slash_tk();
name const & get_plus_tk();
name const & get_star_tk();
name const & get_turnstile_tk();
name const & get_membership_tk();
name const & get_explicit_tk();
name const & get_partial_explicit_tk();
name const & get_at_tk();
name const & get_max_tk();
name const & get_imax_tk();
name const & get_import_tk();
name const & get_prelude_tk();
name const & get_show_tk();
name const & get_have_tk();
name const & get_assume_tk();
name const & get_suppose_tk();
name const & get_fun_tk();
name const & get_match_tk();
name const & get_dotdot_tk();
name const & get_ellipsis_tk();
name const & get_raw_tk();
name const & get_true_tk();
name const & get_false_tk();
name const & get_options_tk();
name const & get_commands_tk();
name const & get_instances_tk();
name const & get_class_tk();
name const & get_classes_tk();
name const & get_attributes_tk();
name const & get_arrow_tk();
name const & get_larrow_tk();
name const & get_hiding_tk();
name const & get_example_tk();
name const & get_exposing_tk();
name const & get_renaming_tk();
name const & get_replacing_tk();
name const & get_extends_tk();
name const & get_as_tk();
name const & get_whnf_tk();
name const & get_in_tk();
name const & get_let_tk();
name const & get_assign_tk();
name const & get_from_tk();
name const & get_then_tk();
name const & get_else_tk();
name const & get_by_tk();
name const & get_rewrite_tk();
name const & get_begin_tk();
name const & get_end_tk();
name const & get_private_tk();
name const & get_protected_tk();
name const & get_inline_tk();
name const & get_definition_tk();
name const & get_meta_tk();
name const & get_mutual_tk();
name const & get_theorem_tk();
name const & get_axiom_tk();
name const & get_axioms_tk();
name const & get_constant_tk();
name const & get_constants_tk();
name const & get_variable_tk();
name const & get_variables_tk();
name const & get_structure_tk();
name const & get_attribute_tk();
name const & get_with_tk();
name const & get_without_tk();
name const & get_prev_tk();
name const & get_scoped_tk();
name const & get_foldr_tk();
name const & get_foldl_tk();
name const & get_binder_tk();
name const & get_binders_tk();
name const & get_precedence_tk();
name const & get_infix_tk();
name const & get_infixl_tk();
name const & get_infixr_tk();
name const & get_postfix_tk();
name const & get_prefix_tk();
name const & get_notation_tk();
name const & get_name_tk();
name const & get_calc_tk();
name const & get_root_tk();
name const & get_fields_tk();
name const & get_trust_tk();
name const & get_inductive_tk();
name const & get_instance_tk();
name const & get_this_tk();
name const & get_noncomputable_tk();
name const & get_exclam_tk();
name const & get_theory_tk();
name const & get_key_equivalences_tk();
name const & get_using_tk();
name const & get_using_well_founded_tk();
}
